import h5py
import numpy as np
import cv2
import math
import json
import sys
from scipy import interpolate
from zebrazoom.code.getImage.getForegroundImage import getForegroundImage
from zebrazoom.code.getImage.headEmbededFrame import headEmbededFrame
from scipy.interpolate import UnivariateSpline
from numpy import linspace

from zebrazoom.code.trackingFolder.tailTrackingFunctionsFolder.tailTrackingExtremityDetectFolder.getMidline import getMidline
from zebrazoom.code.trackingFolder.tailTrackingFunctionsFolder.tailTrackingExtremityDetectFolder.findTailExtremeteFolder.findTailExtremete import findTailExtremete
from zebrazoom.code.trackingFolder.tailTrackingFunctionsFolder.tailTrackingExtremityDetectFolder.findTheTwoSides import findTheTwoSides
from zebrazoom.code.trackingFolder.tailTrackingFunctionsFolder.tailTrackingExtremityDetectFolder.findBodyContour import findBodyContour
from zebrazoom.code.trackingFolder.tailTrackingFunctionsFolder.tailTrackingExtremityDetectFolder.Rotate import Rotate
from zebrazoom.code.trackingFolder.tailTrackingFunctionsFolder.tailTrackingExtremityDetectFolder.findTailExtremeteFolder.checkIfMidlineIsInBlob import checkIfMidlineIsInBlob

# from kalmanFilter import kalman_xy
# from trackingFunctions import calculateAngle
# from trackingFunctions import distBetweenThetas
# from trackingFunctions import assignValueIfBetweenRange

 
def tailTrackingExtremityDetect(headPosition,nbTailPoints,i,thresh1,frame,debugAdv,heading, hyperparameters, initialCurFrame, back):
  
  newHeading = -1
  
  dst = frame.copy()
  dst = cv2.cvtColor(dst, cv2.COLOR_GRAY2RGB)
  firstFrame = hyperparameters["firstFrame"]
  lastFrame = hyperparameters["lastFrame"]
  
  if hyperparameters["debugTrackingThreshImg"]:
    if hyperparameters["debugTrackingPtExtremeLargeVerticals"]:
      cv2.imshow('debugTrackingThreshImg', thresh1[int(headPosition[1])-200:len(thresh1), :] )
    else:
      cv2.imshow('debugTrackingThreshImg', thresh1)
    cv2.waitKey(0)
  
  # Finding blob corresponding to the body of the fish
  bodyContour = findBodyContour(headPosition, hyperparameters, thresh1, initialCurFrame, back)
  if type(bodyContour) != int:
    # Finding the two sides of the fish
    res = findTheTwoSides(headPosition, bodyContour, dst, hyperparameters)
    if len(res) == 3:
      heading = res[2]
      newHeading = res[2]
    # Finding tail extremity
    rotatedContour = bodyContour.copy()
    rotatedContour = Rotate(rotatedContour,int(headPosition[0]),int(headPosition[1]),heading,dst)
    [MostCurvyIndex, distance2] = findTailExtremete(rotatedContour, bodyContour, headPosition[0], int(res[0]), int(res[1]), debugAdv, dst, hyperparameters["tailExtremityMaxJugeDecreaseCoeff"], hyperparameters)
    if debugAdv:
      # Head Center
      cv2.circle(dst, (int(headPosition[0]),int(headPosition[1])), 3, (255, 255, 0), -1)
      # Tail basis 1
      pt1 = bodyContour[int(res[0])][0]
      cv2.circle(dst, (pt1[0],pt1[1]), 3, (255, 0, 0), -1)
      # Tail basis 2
      pt1 = bodyContour[int(res[1])][0]
      cv2.circle(dst, (pt1[0],pt1[1]), 3, (180, 0, 0), -1)
      # Tail extremity
      pt1 = bodyContour[int(MostCurvyIndex)][0]
      cv2.circle(dst, (pt1[0],pt1[1]), 3, (0, 0, 255), -1)
      #
      if hyperparameters["debugTrackingPtExtremeLargeVerticals"]:
        dst = dst[int(headPosition[1])-200:len(dst), :]
      # Plotting points
      cv2.imshow('Frame', dst)
      cv2.waitKey(0)
    
    # Getting Midline
    if hyperparameters["detectMouthInsteadOfHeadTwoSides"] == 0:
      tail = getMidline(int(res[0]), int(res[1]), int(MostCurvyIndex), bodyContour, dst, nbTailPoints-1, distance2, debugAdv, hyperparameters, nbTailPoints)
    else:
      tail = getMidline(int(res[0]), int(res[1]), int(MostCurvyIndex), bodyContour, dst, nbTailPoints, distance2, debugAdv, hyperparameters, nbTailPoints)
      tail = np.array([tail[0][1:len(tail[0])]])
    
    if False:
      maxDistContourToTail = -1
      for contourPt in bodyContour:
        contourPtX = contourPt[0][0]
        contourPtY = contourPt[0][1]
        minDistContourPointToTail = 1000000000000
        for tailPt in np.append(tail[0], np.array([headPosition]), axis=0):
          tailPtX = tailPt[0]
          tailPtY = tailPt[1]
          dist = math.sqrt((tailPtX - contourPtX)**2 + (tailPtY - contourPtY)**2)
          if dist < minDistContourPointToTail:
            minDistContourPointToTail = dist
        if minDistContourPointToTail > maxDistContourToTail:
          maxDistContourToTail = minDistContourPointToTail
      print("maxDistContourToTail:", maxDistContourToTail, "; tailLength:", hyperparameters["minTailSize"]*10)
      
    if False:
      for pt in bodyContour:
        cv2.circle(dst, (pt[0][0], pt[0][1]), 3, (0, 0, 255), -1)
      cv2.imshow('Frame', dst)
      cv2.waitKey(0)
    
    # Optimizing midline if necessary
    midlineIsInBlobTrackingOptimization = hyperparameters["midlineIsInBlobTrackingOptimization"]
    if midlineIsInBlobTrackingOptimization:
      [allInside, tailLength] = checkIfMidlineIsInBlob(int(res[0]), int(res[1]), int(MostCurvyIndex), bodyContour, dst, nbTailPoints-1, distance2, debugAdv, hyperparameters, nbTailPoints)
      if allInside == False:
        n = len(bodyContour)
        maxTailLength = -1
        for j in range(0, n):
          [allInside, tailLength] = checkIfMidlineIsInBlob(int(res[0]), int(res[1]), j, bodyContour, dst, nbTailPoints-1, distance2, debugAdv, hyperparameters, nbTailPoints)
          if allInside:
            if tailLength > maxTailLength:
              MostCurvyIndex = j
              maxTailLength = tailLength
        tail = getMidline(int(res[0]), int(res[1]), int(MostCurvyIndex), bodyContour, dst, nbTailPoints-1, distance2, debugAdv, hyperparameters, nbTailPoints)
    # Applying snake on tail
    applySnake = False
    if applySnake:
      tail2 = tail[0]
      n = len(tail2)
      # tail2[n-1][0] = tail2[n-1][0] + (tail2[n-1][0] - tail2[n-2][0]) * 6
      # tail2[n-1][1] = tail2[n-1][1] + (tail2[n-1][1] - tail2[n-2][1]) * 6
      # print(type(tail))
      # r = np.linspace(tail2[0][0], tail2[0][0] + (tail2[1][0]-tail2[0][0]) * 15, 9)
      # c = np.linspace(tail2[0][1], tail2[0][1] + (tail2[1][1]-tail2[0][1]) * 15, 9)
      # tail2 = np.array([r, c]).T
      # r = np.linspace(tail2[0][0], tail2[n-1][0], 9)
      # c = np.linspace(tail2[0][1], tail2[n-1][1], 9)
      # tail2 = np.array([r, c]).T
      from skimage.color import rgb2gray
      from skimage.filters import gaussian
      from skimage.segmentation import active_contour
      snake = active_contour(gaussian(frame, 3), tail2, w_edge=-1000, bc="fixed")
      # snake = active_contour(gaussian(frame, 3), tail2, w_edge=0, bc="fixed-free")
      print(snake)
      # snake = tail2
      tail[0] = snake
  
  else:
    
    tail = np.zeros((1, 0, 2))
    point = np.array([0, 0])
    for i in range(0, nbTailPoints):
      tail = np.insert(tail, 0, point, axis=1)  
    # if hyperparameters["detectMouthInsteadOfHeadTwoSides"] != 0:
      # tail = np.insert(tail, 0, point, axis=1)  
  
  # Inserting head position, smoothing tail and creating output
  # if hyperparameters["detectMouthInsteadOfHeadTwoSides"] == 0:
    # tail = np.insert(tail, 0, headPosition, axis=1)
  tail = np.insert(tail, 0, headPosition, axis=1)
 
  if nbTailPoints != len(tail[0]):
    print("small problem 1 in tailTrackingExtremityDetect")
  
  # output = np.zeros((1, len(tail[0]), 2))
  output = np.zeros((1, nbTailPoints, 2))
  
  for idx, x in enumerate(tail[0]):
    if idx < nbTailPoints:
      output[0][idx][0] = x[0]
      output[0][idx][1] = x[1]
    else:
      print("small problem 2 in tailTrackingExtremityDetect")

  return [output, newHeading]



import numpy as np
import json
import cv2
import math
from zebrazoom.code.getHyperparameters import getHyperparametersSimple
from zebrazoom.code.getImage.getForegroundImage import getForegroundImage
from zebrazoom.mainZZ import mainZZ
import cvui
import pickle
import json
import os
import re
from zebrazoom.code.getBackground import getBackground
from zebrazoom.code.findWells import findWells
from zebrazoom.code.trackingFolder.tracking import tracking
from zebrazoom.code.GUI.adjustParameterInsideAlgoFunctions import prepareConfigFileForParamsAdjustements

from zebrazoom.code.GUI.automaticallyFindOptimalParametersFunctions import getGroundTruthFromUser, findBestBackgroundSubstractionParameterForEachImage, findInitialBlobArea, boutDetectionParameters

def automaticallyFindOptimalParameters(self, controller, realExecThroughGUI, detectBouts, method, nbOfAnimalToTrackNotConstantOverTime, adjustBackgroundExtractionBasedOnNumberOfBlackPixels):
  
  nbOfImagesToManuallyClassify = 3
  saveIntermediary = False # Should be set to False except when debugging
  zebrafishToTrack = True
  
  # # # Getting ground truth from user: head center and tail extremity coordinates for a few frames
  
  if realExecThroughGUI:
    zebrafishToTrack = (self.organism == 'zebrafishNew')
    [initialConfigFile, videoPath, data, wellPositions, pathToVideo, videoNameWithExt, videoName, videoExt] = getGroundTruthFromUser(self, controller, nbOfImagesToManuallyClassify, saveIntermediary, zebrafishToTrack)
  else:
    toSave = pickle.load(open(self, 'rb'))
    initialConfigFile = toSave[0]
    videoPath         = toSave[1]
    data              = toSave[2]
    wellPositions     = toSave[3]
    pathToVideo       = toSave[4]
    videoNameWithExt  = toSave[5]
    videoName         = toSave[6]
    videoExt          = toSave[7]
    zebrafishToTrack  = toSave[8]
  
  print("initialConfigFile:", initialConfigFile)
  
  # # # Starting the process of finding the best hyperparameters to track the video
  
  if zebrafishToTrack:
    configFile = {"extractAdvanceZebraParameters": 0, "headEmbeded": 0, "nbWells": 1, "noBoutsDetection": 1, "noChecksForBoutSelectionInExtractParams": 1, "trackingPointSizeDisplay": 4, "validationVideoPlotHeading": 1, "nbAnimalsPerWell": 1, "forceBlobMethodForHeadTracking": 1, "multipleHeadTrackingIterativelyRelaxAreaCriteria": 1, "erodeIter":0, "minArea": 0, "maxArea": 100000000000000, "minAreaBody": 0, "maxAreaBody": 100000000000000, "headSize": 20, "minTailSize": 0, "maxTailSize": 100000000000000, "paramGaussianBlur": 25, "extractBackWhiteBackground": 1, "dilateIter": 0, "thresholdForBlobImg": 254, "findContourPrecision": "CHAIN_APPROX_NONE", "midlineIsInBlobTrackingOptimization": 0, "checkAllContourForTailExtremityDetect": 1, "recalculateForegroundImageBasedOnBodyArea": 0, "headingCalculationMethod": "simplyFromPreviousCalculations", "detectMouthInsteadOfHeadTwoSides": 1, "findCenterOfAnimalByIterativelyDilating": 1}
  else:
    configFile = {"nbWells": 1, "noBoutsDetection": 1, "trackingPointSizeDisplay": 4, "validationVideoPlotHeading": 0, "nbAnimalsPerWell": 1, "forceBlobMethodForHeadTracking": 1, "multipleHeadTrackingIterativelyRelaxAreaCriteria": 1, "erodeIter":0, "minArea": 0, "maxArea": 100000000000000, "minAreaBody": 0, "maxAreaBody": 100000000000000, "headSize": 20, "minTailSize": 0, "maxTailSize": 100000000000000, "paramGaussianBlur": 25, "extractBackWhiteBackground": 1, "dilateIter": 0, "thresholdForBlobImg": 254, "findContourPrecision": "CHAIN_APPROX_NONE", "recalculateForegroundImageBasedOnBodyArea": 0, "headingCalculationMethod": "simplyFromPreviousCalculations", "trackTail": 0}
  
  if "firstFrame" in initialConfigFile:
    configFile["firstFrame"] = initialConfigFile["firstFrame"]
  if "lastFrame" in initialConfigFile:
    configFile["lastFrame"]  = initialConfigFile["lastFrame"]
  
  hyperparameters = getHyperparametersSimple(configFile)
  
  background = getBackground(os.path.join(pathToVideo, videoNameWithExt), hyperparameters)
  
  hyperparameters["minPixelDiffForBackExtract"] = 20
  
  # # # For each frame that has some ground truth provided by the user:
  # # # Finds the parameter minPixelDiffForBackExtract (as well as the corresponding number of black pixels in the black frame) that leads to a tail extremity detected as close as possible to the tail extremity value provided by the user
  
  data = findBestBackgroundSubstractionParameterForEachImage(data, videoPath, background, wellPositions, hyperparameters, videoName, zebrafishToTrack)
  
  # # # Choosing the hyperparameters related to the background substraction and to the tail length as well as the headSize hyperparameter (for the heading calculation)
  
  configFile["nbAnimalsPerWell"]      = initialConfigFile["nbAnimalsPerWell"]
  hyperparameters["nbAnimalsPerWell"] = initialConfigFile["nbAnimalsPerWell"]
  
  maxTailLengthManual = 100000000000
  if zebrafishToTrack:
    tailLengthManualOptions = []
    for image in data:
      tailLengthManualOptions.append(image["tailLengthManual"])
    maxTailLengthManual = max(tailLengthManualOptions)
    print("tailLengthManualOptions:", tailLengthManualOptions)
    print("maxTailLengthManual:", maxTailLengthManual)
  
  bestMinPixelDiffForBackExtract = -1
  bestMinPixelDiffForBackExtractOptions = []
  bodyContourAreaOptions = []
  tailLengthOptions = []
  for image in data:
    if image["lowestTailTipDistError"] != 1000000000 and (not("tailLength" in image) or (image["tailLength"] < 10 * maxTailLengthManual)):
      if image["bodyContourArea"] != -1:
        bestMinPixelDiffForBackExtractOptions.append(image["bestMinPixelDiffForBackExtract"])
        bodyContourAreaOptions.append(image["bodyContourArea"])
        if zebrafishToTrack:
          tailLengthOptions.append(image["tailLength"])
  
  print("bestMinPixelDiffForBackExtractOptions:", bestMinPixelDiffForBackExtractOptions)
  print("tailLengthOptions:", tailLengthOptions)
  print("bodyContourAreaOptions:", bodyContourAreaOptions)
  
  if len(bestMinPixelDiffForBackExtractOptions):
    ind = np.argmin(bodyContourAreaOptions)
    bestMinPixelDiffForBackExtract = bestMinPixelDiffForBackExtractOptions[ind] # THIS IS THE MOST IMPORTANT PART
    bodyContourArea = bodyContourAreaOptions[ind]
    if zebrafishToTrack:
      tailLength = tailLengthOptions[ind]
  else:
    print("NEED TO START OVER: bodyContourArea")
    bodyContourArea = 0
  if bestMinPixelDiffForBackExtract == -1:
    print("NEED TO START OVER")
  
  hyperparameters["minPixelDiffForBackExtract"] = bestMinPixelDiffForBackExtract
  if nbOfAnimalToTrackNotConstantOverTime:
    hyperparameters["multipleHeadTrackingIterativelyRelaxAreaCriteria"] = 0
  else:
    if adjustBackgroundExtractionBasedOnNumberOfBlackPixels:
      hyperparameters["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"] = bodyContourArea * configFile["nbAnimalsPerWell"]
  hyperparameters["maxAreaBody"] = 5 * bodyContourArea
  hyperparameters["minAreaBody"] = int(bodyContourArea / 5)
  hyperparameters["trackingPointSizeDisplay"] = int(np.ceil(math.sqrt(bodyContourArea) / 15))
  if zebrafishToTrack:
    hyperparameters["minTailSize"] = tailLength / 10
    hyperparameters["maxTailSize"] = tailLength * 2
    hyperparameters["headSize"]    = int(tailLength / 2)
  
  configFile["minPixelDiffForBackExtract"] = bestMinPixelDiffForBackExtract
  if nbOfAnimalToTrackNotConstantOverTime:
    configFile["multipleHeadTrackingIterativelyRelaxAreaCriteria"] = 0
  else:
    if adjustBackgroundExtractionBasedOnNumberOfBlackPixels:
      configFile["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"] = bodyContourArea * configFile["nbAnimalsPerWell"]
  configFile["maxAreaBody"] = 5 * bodyContourArea
  configFile["minAreaBody"] = int(bodyContourArea / 5)
  configFile["trackingPointSizeDisplay"] = int(np.ceil(math.sqrt(bodyContourArea) / 15))
  if zebrafishToTrack:
    configFile["minTailSize"] = tailLength / 10
    configFile["maxTailSize"] = tailLength * 2
    configFile["headSize"]    = int(tailLength / 2)
  
  # # # For each frame that has some ground truth provided by the user: finds the contour clicked by the user and its area USING THE NEW CONFIG FILE PARAMETERS SET JUST ABOVE
  
  data = findInitialBlobArea(data, videoPath, background, wellPositions, hyperparameters, maxTailLengthManual)
  
  # # # Choosing the hyperparameters related to the max and min of the "head" contour (the initial contours found during the tracking)
  
  headContourAreaOptions = []
  for image in data:
    if image["lowestTailTipDistError"] != 1000000000 and (not("tailLength" in image) or (image["tailLength"] < 10 * maxTailLengthManual)):
      if "headContourArea" in image:
        headContourAreaOptions.append(image["headContourArea"])
  print("headContourAreaOptions:", headContourAreaOptions)
  if len(headContourAreaOptions):
    headContourArea = np.mean(headContourAreaOptions)
  else:
    print("Start Over!!!")
  if nbOfAnimalToTrackNotConstantOverTime:
    hyperparameters["minArea"] = 0.5 * headContourArea
    hyperparameters["maxArea"] = 1.5 * headContourArea
    configFile["minArea"] = 0.5 * headContourArea
    configFile["maxArea"] = 1.5 * headContourArea  
  else:
    hyperparameters["minArea"] = 0.7 * headContourArea
    hyperparameters["maxArea"] = 1.3 * headContourArea
    configFile["minArea"] = 0.7 * headContourArea
    configFile["maxArea"] = 1.3 * headContourArea
    
  if not((0 < configFile["minAreaBody"]) and (configFile["minAreaBody"] < configFile["minArea"]) and (configFile["minArea"] < configFile["maxArea"]) and (configFile["maxArea"] < configFile["maxAreaBody"])):
    configFile["minArea"] = configFile["minAreaBody"]
    configFile["maxArea"] = configFile["maxAreaBody"]
  
  hyperparameters["dilateIter"] = 0
  configFile["dilateIter"] = 0
  
  # # # Overwritting some of the hyperparameters from hyperparameters in the initial configFile
  
  listOfParametersToOverwrite = ["extractAdvanceZebraParameters", "nbWells", "nbRowsOfWells", "nbWellsPerRows", "minWellDistanceForWellDetection", "wellOutputVideoDiameter", "wellsAreRectangles", "rectangleWellAreaImageThreshold", "rectangleWellErodeDilateKernelSize", "findRectangleWellArea", "rectangularWellsInvertBlackWhite", "noWellDetection", "oneWellManuallyChosenTopLeft", "oneWellManuallyChosenBottomRight", "multipleROIsDefinedDuringExecution"]
  for parameter in listOfParametersToOverwrite:
    if parameter in initialConfigFile:
      configFile[parameter] = initialConfigFile[parameter]
  
  print("Intermediary config file:", configFile)
    
  # Adjusting config file hyperparameters related to bouts detection
  if (configFile["nbAnimalsPerWell"] == 1) and detectBouts:
    configFile = boutDetectionParameters(data, configFile, pathToVideo, videoName, videoExt, wellPositions, videoPath)
  
  # Setting recalculateForegroundImageBasedOnBodyArea to 1 when asked by user to try to obtain better tail tracking
  if adjustBackgroundExtractionBasedOnNumberOfBlackPixels and method:
    configFile["recalculateForegroundImageBasedOnBodyArea"] = 1
  
  # # # Moving on to the next step
  
  if realExecThroughGUI:
    self.configFile = configFile
    controller.show_frame("FinishConfig")
  
  print("final Config File", configFile)
  
  reference = videoName + '_config.json'
  with open(reference, 'w') as outfile:
    json.dump(configFile, outfile)
  
  


import tkinter as tk
from tkinter import Tk, mainloop, TOP
from threading import Timer

from vars import getGlobalVariables
globalVariables = getGlobalVariables()

def initialise():
  if globalVariables["mac"] != 1:
    f = open("trace.txt","w+")
    f.write("")
    f.close() 
    def launchStuff():
      root = Tk()
      label2 = tk.Text(root)
      label2.pack()
      def changeLabelValue():
        f = open("trace.txt", "r")
        if f.mode == 'r':
          contents = f.read()
        f.close()
        label2.delete('1.0', tk.END)
        label2.insert(tk.END, contents)
        if "ZebraZoom Analysis all finished" in contents:
          # if closePopUpWindowAtTheEnd:
          root.destroy()
        else:
          r = Timer(1.0, changeLabelValue, ())
          r.start()          
      r = Timer(1.0, changeLabelValue, ())
      r.start()
      root.mainloop()
    
    f = Timer(0.5, launchStuff, ())
    f.start()

def prepend(text):
  if globalVariables["mac"] != 1:
    with open("trace.txt", "r+") as f:
      content = f.read()
      f.seek(0, 0)
      f.write(text.rstrip('\r\n') + '\n' + content)
